﻿//////////////////////////////////////////////////////////////////////////////
///   Namespace:        PottsSoftware
///   Class:            SSELibrary
///   Description:      .Net class to wrap the calls from safe managed code to
///                     the unsafe unmanaged dll.
///   Author:           George Potts
///   Initial Date:     2015/08/13
///   Notes:
///
///   Revision History:
//////////////////////////////////////////////////////////////////////////////
using System;
using System.Runtime.InteropServices;

namespace PottsSoftware
{
    partial class SSELibrary : IDisposable
    {
        // 32 bytes => 256 bits for AVX support.
        private int _alignment = 32;

        private ulong _bytes = 0;
        private IntPtr _rawPtr1 = IntPtr.Zero;
        private IntPtr _rawPtr2 = IntPtr.Zero;
        private IntPtr _rawPtr3 = IntPtr.Zero;
        private IntPtr _Ptr1 = IntPtr.Zero;
        private IntPtr _Ptr2 = IntPtr.Zero;
        private IntPtr _Ptr3 = IntPtr.Zero;

        private bool disposed = false;

        public delegate void SSEDelegate1(IntPtr buffer, ulong bytes);
        public delegate void SSEDelegate2(IntPtr buffer, ulong bytes, int operation);
        public delegate void SSEDelegate3(IntPtr dest, IntPtr src, ulong bytes);
        public delegate void SSEDelegate4(IntPtr dest, IntPtr src, IntPtr mask, ulong bytes);
        public delegate void SSEDelegate5(IntPtr dest, IntPtr src, ulong bytes, int step);
        public delegate void SSEDelegate6(IntPtr dest, IntPtr wave, IntPtr src, ulong bytes, int step);

        public ulong Bytes => _bytes;

        #region Lifecycle
        public SSELibrary(ulong bytes)
        {
            Allocate(bytes);
        }

        ~SSELibrary()
        {
            Dispose(false);
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        public virtual void Dispose(bool disposing)
        {
            if (!disposed)
            {
                disposed = true;
            }
        }
        #endregion // Lifecycle

        // Call to determine if we have any dll loading issues.

        public static bool IsPresent()
        {
            bool fPresent = false;
            try
            {
                fPresent = SafeNativeMethods.IsPresent();
            }
            catch (Exception ) { ; }
            //catch (Exception ) { Console.WriteLine("IsPresent: exception"); }

            return fPresent;
        }

        #region Allocation
        private void Allocate(ulong bytes)
        {
            if (bytes < 1)
            {
                throw new Exception(string.Format("SSELibrary::Allocate: Attempt to Allocate invalid number of bytes: {0}", bytes));
            }

            if (_bytes < bytes)
            {
                DeAllocate();
                _bytes = bytes;

                // Overallocate to make sure we're aligned with our
                // pointers and have extra padding at the end for odd buffers.
                _rawPtr1 = Marshal.AllocHGlobal((int)_bytes + (2*_alignment));
                _rawPtr2 = Marshal.AllocHGlobal((int)_bytes + (2*_alignment));
                _rawPtr3 = Marshal.AllocHGlobal((int)_bytes + (2*_alignment));
                _Ptr1 = new IntPtr(_alignment * (((long)_rawPtr1 + (_alignment -1)) / _alignment));
                _Ptr2 = new IntPtr(_alignment * (((long)_rawPtr2 + (_alignment -1)) / _alignment));
                _Ptr3 = new IntPtr(_alignment * (((long)_rawPtr2 + (_alignment -1)) / _alignment));
            }
        }

        private void DeAllocate()
        {
            if (_rawPtr1 != IntPtr.Zero)
            {
                Marshal.FreeHGlobal(_rawPtr1);
                _rawPtr1 = IntPtr.Zero;
                _Ptr1 = IntPtr.Zero;
            }
            if (_rawPtr2 != IntPtr.Zero)
            {
                Marshal.FreeHGlobal(_rawPtr2);
                _rawPtr2 = IntPtr.Zero;
                _Ptr2 = IntPtr.Zero;
            }
            if (_rawPtr3 != IntPtr.Zero)
            {
                Marshal.FreeHGlobal(_rawPtr3);
                _rawPtr3 = IntPtr.Zero;
                _Ptr3 = IntPtr.Zero;
            }
        }
        #endregion // Allocation

        #region Delegate_Wrappers
        private void Operation1(byte[] src, ulong bytes, SSEDelegate1 dlg)
        {
            Allocate(bytes);
            unsafe
            {
                fixed (byte* srcPtr = src)
                {
                    Marshal.Copy(src, 0, _Ptr1, (int)bytes);
                    dlg(_Ptr1,  bytes);
                    Marshal.Copy(_Ptr1, src, 0, (int)bytes);
                }
            }
        }

        private void Operation2(byte[] dest, byte[] src, ulong bytes, SSEDelegate3 dlg)
        {
            Allocate(bytes);
            unsafe
            {
                fixed (byte* destPtr = dest, srcPtr = src)
                {
                    Marshal.Copy(dest, 0, _Ptr1, (int)bytes);
                    Marshal.Copy(src, 0, _Ptr2, (int)bytes);
                    dlg(_Ptr1, _Ptr2, bytes);
                    Marshal.Copy(_Ptr1, dest, 0, (int)bytes);
                }
            }
        }

        private void Operation3(byte[] dest, byte[] src, byte[] mask, ulong bytes, SSEDelegate4 dlg)
        {
            Allocate(bytes);
            unsafe
            {
                fixed (byte* destPtr = dest, srcPtr = src, maskPtr = mask)
                {
                    Marshal.Copy(dest, 0, _Ptr1, (int)bytes);
                    Marshal.Copy(src, 0, _Ptr2, (int)bytes);
                    Marshal.Copy(mask, 0, _Ptr3, (int)bytes);
                    dlg(_Ptr1, _Ptr2, _Ptr3, bytes);
                    Marshal.Copy(_Ptr1, dest, 0, (int)bytes);
                }
            }
        }
#endregion // Delegate_Wrappers
    }
}