//////////////////////////////////////////////////////////////////////////////
//   Author:       George Potts
//   Copyright:    PottsSoftware
//   Initial Date: 2015/08/13
// 
//   Description:  Detect CPU features.
//
//   Notes:
//      A lot of the information below was pulled from here -
//      http://www.flounder.com/cpuid_explorer2.htm
//      I think most of the original flags were copied from here:
//      http://datasheets.chipdb.org/Intel/x86/CPUID/24161821.pdf
//      Additional information can be found on Wikipedia for CPUID -
//      https://en.wikipedia.org/wiki/CPUID
//
//      TODO: Verify that they're also valid for AMD chips. As of now Intel is more complete.
//
//   Revision History:
//
//////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include <immintrin.h>
#include <intrin.h>
#include "SSELibrary.h"

typedef union {
    struct {                      // low order
        UINT SteppingID:4;         // 3..0
        UINT ModelID:4;            // 7..4
        UINT FamilyID:4;           // 11..8
        UINT ProcessorType:2;      // 13..12
        UINT Reserved2:2;          // 15..14
        UINT ExtendedModel:4;      // 19..16
        UINT ExtendedFamily:8;     // 27..20
        UINT Reserved:4;           // 31..28
    } Intel;                      // high order

    struct {                      // low order
        UINT SteppingID:4;         // 3..0
        UINT ModelID:4;            // 7..4
        UINT FamilyID:4;           // 11..8
        UINT Reserved2:4;          // 15..12
        UINT ExtendedModel:4;      // 19..16
        UINT ExtendedFamily:8;     // 27..20
        UINT Reserved:4;           // 31..28
    } AMD;                       // high order
    UINT w;
} EAXF1;

typedef union {
    struct { // low order
        UINT BrandIndex:8;
        UINT CLFLUSHLineSize:8;
        UINT LogicalProcessors:8;
        UINT InitialAPICID:8;
    } bits;  // high order
    UINT w;
} EBXF1;

typedef union {
    struct { // low order
        UINT SSE3:1;                 // 0
        UINT PCLMULQDQ:1;            // 1
        UINT DTES64:1;               // 2
        UINT MWAIT:1;                // 3   // aka MONITOR
        UINT CPL:1;                  // 4   
        UINT VMX:1;                  // 5   
        UINT SMX:1;                  // 6   
        UINT EST:1;                  // 7   
        UINT TM2:1;                  // 8   
        UINT SSSE3:1;                // 9   
        UINT L1:1;                   // 10  
        UINT SDBG:1;                 // 11  
        UINT FMA:1;                  // 12 
        UINT CMPXCHG16B:1;           // 13
        UINT XTPR:1;                 // 14
        UINT PDCM:1;                 // 15
        UINT Reserved6:1;            // 16
        UINT PCID:1;                 // 17
        UINT DCA:1;                  // 18
        UINT SSE41:1;                // 19
        UINT SSE42:1;                // 20
        UINT X2APIC:1;               // 21
        UINT MOVBE:1;                // 22
        UINT POPCNT:1;               // 23
        UINT TSCDEADLINE:1;          // 24
        UINT AES:1;                  // 25
        UINT XSAVE:1;                // 26
        UINT OSXSAVE:1;              // 27
        UINT AVX:1;                  // 28
        UINT F16C:1;                 // 29
        UINT RDRAND:1;               // 30
        UINT HYPERVISOR:1;           // 31
    } Intel;  // high order

    struct { // low order
        UINT SSE3:1;                 // 0     
        UINT Reserved1:3;            // 3..1 
        UINT Reserved2:4;            // 7..4 
        UINT Reserved3:4;            // 11..8
        UINT FMA:1;                  // 12  
        UINT CMPXCHG16B:1;           // 13
        UINT Reserved6:5;            // 18..14
        UINT SSE41:1;                // 19
        UINT SSE42:1;                // 20
        UINT Reserved7:18;           // 31..21
    } AMD;  // high order
    UINT w;
} ECXF1;

typedef union {
    struct { // low order
        UINT FPU:1;                  // 0
        UINT VME:1;                  // 1
        UINT DE:1;                   // 2
        UINT PSE:1;                  // 3
        UINT TSC:1;                  // 4
        UINT MSR:1;                  // 5
        UINT PAE:1;                  // 6
        UINT MCE:1;                  // 7
        UINT CX8:1;                  // 8
        UINT APIC:1;                 // 9
        UINT reserved1:1;            // 10
        UINT SEP:1;                  // 11
        UINT MTRR:1;                 // 12
        UINT PGE:1;                  // 13
        UINT MCA:1;                  // 14
        UINT CMOV:1;                 // 15
        UINT PAT:1;                  // 16
        UINT PSE36:1;                // 17     
        UINT PSN:1;                  // 18
        UINT CFLSH:1;                // 19     
        UINT Reserved2:1;            // 20
        UINT DS:1;                   // 21
        UINT ACPI:1;                 // 22
        UINT MMX:1;                  // 23
        UINT FXSR:1;                 // 24
        UINT SSE:1;                  // 25
        UINT SSE2:1;                 // 26     
        UINT SS:1;                   // 27
        UINT HTT:1;                  // 28     
        UINT TM:1;                   // 29
        UINT IA64:1;                 // 30
        UINT PBE:1;                  // 31
    } Intel;  // high order

    struct { // low order
        UINT FPU:1;                  // 0
        UINT VME:1;                  // 1
        UINT DE:1;                   // 2
        UINT PSE:1;                  // 3
        UINT MSR:1;                  // 4
        UINT TSC:1;                  // 5
        UINT PAE:1;                  // 6
        UINT MCE:1;                  // 7
        UINT CX8:1;                  // 8
        UINT APIC:1;                 // 9
        UINT Reserved1:1;            // 10
        UINT SEP:1;                  // 11
        UINT MTRR:1;                 // 12
        UINT PGE:1;                  // 13
        UINT MCA:1;                  // 14
        UINT CMOV:1;                 // 15
        UINT PAT:1;                  // 16
        UINT PSE36:1;                // 17
        UINT Reserved5:1;            // 18
        UINT CFLSH:1;                // 19     
        UINT Reserved2:3;            // 22..20
        UINT MMX:1;                  // 23
        UINT FXSR:1;                 // 24
        UINT SSE:1;                  // 25
        UINT SSE2:1;                 // 26     
        UINT Reserved4:1;            // 27
        UINT HTT:1;                  // 28     
        UINT Reserved3:3;            // 31..29
    } AMD;  // high order
    UINT w;
} EDXF1;

typedef union {
    struct { // low order
        UINT FSGSBASE:1;            // 0
        UINT Reserved1:1;           // 1
        UINT Reserved2:1;           // 2
        UINT BMI1:1;                // 3
        UINT HLE:1;                 // 4
        UINT AVX2:1;                // 5
        UINT Reserved3:1;           // 6
        UINT SMEP:1;                // 7
        UINT BMI2:1;                // 8
        UINT ERMS:1;                // 9
        UINT INVPCID:1;             // 10
        UINT RTM:1;                 // 11
        UINT Reserved4:1;           // 12
        UINT Reserved5:1;           // 13
        UINT MPX:1;                 // 14
        UINT Reserved6:1;           // 15
        UINT AVX512F:1;             // 16
        UINT AVX512DQ:1;            // 17
        UINT RDSEED:1;              // 18
        UINT ADX:1;                 // 19
        UINT SMAP:1;                // 20
        UINT AVX512IFMA:1;          // 21
        UINT PCOMMIT:1;             // 22
        UINT CLFLUSHOPT:1;          // 23
        UINT CLWB:1;                // 24
        UINT IPT:1;                 // 25
        UINT AVX512PF:1;            // 26
        UINT AVX512ER:1;            // 27
        UINT AVX512CD:1;            // 28
        UINT SHA:1;                 // 29
        UINT AVX512BW:1;            // 30
        UINT AVX512VL:1;            // 31
    } Intel;  // high order

    struct { // low order
        UINT FSGSBASE:1;            // 0
        UINT Reserved1:1;           // 1
        UINT Reserved2:1;           // 2
        UINT BMI1:1;                // 3
        UINT HLE:1;                 // 4
        UINT AVX2:1;                // 5
        UINT Reserved3:1;           // 6
        UINT SMEP:1;                // 7
        UINT BMI2:1;                // 8
        UINT ERMS:1;                // 9
        UINT INVPCID:1;             // 10
        UINT RTM:1;                 // 11
        UINT Reserved4:1;           // 12
        UINT Reserved5:1;           // 13
        UINT MPX:1;                 // 14
        UINT Reserved6:1;           // 15
        UINT AVX512F:1;             // 16
        UINT AVX512DQ:1;            // 17
        UINT RDSEED:1;              // 18
        UINT ADX:1;                 // 19
        UINT SMAP:1;                // 20
        UINT AVX512IFMA:1;          // 21
        UINT PCOMMIT:1;             // 22
        UINT CLFLUSHOPT:1;          // 23
        UINT CLWB:1;                // 24
        UINT IPT:1;                 // 25
        UINT AVX512PF:1;            // 26
        UINT AVX512ER:1;            // 27
        UINT AVX512CD:1;            // 28
        UINT SHA:1;                 // 29
        UINT AVX512BW:1;            // 30
        UINT AVX512VL:1;            // 31
    } AMD;  // high order
    UINT w;
} EBXF7;

typedef union {
    struct { // low order
        UINT PREFETCHWT1:1;          // 0
        UINT AVX512VBMI:1;           // 1
        UINT reserved1:30;           // 31-2
    } Intel;  // high order

    struct { // low order
        UINT PREFETCHWT1:1;          // 0
        UINT AVX512VBMI:1;           // 1
        UINT reserved1:30;           // 31-2
    } AMD;  // high order
    UINT w;
} ECXF7;

typedef union {
    struct { // low order
        UINT LAHF:1;                // 0
        UINT CMPLEGACY:1;           // 1
        UINT SVM:1;                 // 2
        UINT EXTAPIC:1;             // 3
        UINT CR8LEGACY:1;           // 4
        UINT ABM:1;                 // 5
        UINT SSE4A:1;               // 6
        UINT MISALIGNSSE:1;         // 7
        UINT _3DNOWPREFETCH:1;      // 8
        UINT OSVW:1;                // 9
        UINT IBS:1;                 // 10
        UINT XOP:1;                 // 11
        UINT SKINIT:1;              // 12
        UINT WDT:1;                 // 13
        UINT Reserved1:1;           // 14
        UINT LWP:1;                 // 15
        UINT FMA4:1;                // 16
        UINT TCE:1;                 // 17
        UINT Reserved2:2;           // 18
        UINT NODEIDMSR:1;           // 19
        UINT Reserved3:1;           // 20
        UINT TBM:1;                 // 21
        UINT TOPOEXT:1;             // 22
        UINT PERFCTRCORE:1;         // 23
        UINT PERFCTRNB:1;           // 24
        UINT Reserved4:1;           // 25
        UINT DBX:1;                 // 26
        UINT PERFTSC:1;             // 27
        UINT PCXL2I:1;              // 28
        UINT Reserved5:1;           // 29
        UINT Reserved6:1;           // 30
        UINT Reserved7:1;           // 31
    } Intel;  // high order

    struct { // low order
        UINT LAHF:1;                // 0
        UINT CMPLEGACY:1;           // 1
        UINT SVM:1;                 // 2
        UINT EXTAPIC:1;             // 3
        UINT CR8LEGACY:1;           // 4
        UINT ABM:1;                 // 5
        UINT SSE4A:1;               // 6
        UINT MISALIGNSSE:1;         // 7
        UINT _3DNOWPREFETCH:1;      // 8
        UINT OSVW:1;                // 9
        UINT IBS:1;                 // 10
        UINT XOP:1;                 // 11
        UINT SKINIT:1;              // 12
        UINT WDT:1;                 // 13
        UINT Reserved1:1;           // 14
        UINT LWP:1;                 // 15
        UINT FMA4:1;                // 16
        UINT TCE:1;                 // 17
        UINT Reserved2:2;           // 18
        UINT NODEIDMSR:1;           // 19
        UINT Reserved3:1;           // 20
        UINT TBM:1;                 // 21
        UINT TOPOEXT:1;             // 22
        UINT PERFCTRCORE:1;         // 23
        UINT PERFCTRNB:1;           // 24
        UINT Reserved4:1;           // 25
        UINT DBX:1;                 // 26
        UINT PERFTSC:1;             // 27
        UINT PCXL2I:1;              // 28
        UINT Reserved5:1;           // 29
        UINT Reserved6:1;           // 30
        UINT Reserved7:1;           // 31
    } AMD;  // high order
    UINT w;
} ECXF8;

typedef union {
    struct { // low order
        UINT FPU:1;                 // 0
        UINT VME:1;                 // 1
        UINT DE:1;                  // 2
        UINT PSE:1;                 // 3
        UINT TSC:1;                 // 4
        UINT MSR:1;                 // 5
        UINT PAE:1;                 // 6
        UINT MCE:1;                 // 7
        UINT CX8:1;                 // 8
        UINT APIC:1;                // 9
        UINT Reserved1:1;           // 10
        UINT SYCALL:1;              // 11
        UINT MTRR:1;                // 12
        UINT PGE:1;                 // 13
        UINT MCA:1;                 // 14
        UINT CMOV:1;                // 15
        UINT PAT:1;                 // 17
        UINT PSE36:1;               // 17
        UINT Reserved2:1;           // 18
        UINT MP:1;                  // 19
        UINT NX:1;                  // 20
        UINT Reserved3:1;           // 21
        UINT MMXEXT:1;              // 22
        UINT MMX:1;                 // 23
        UINT FXSR:1;                // 24
        UINT FXSROPT:1;             // 25
        UINT PDPE1GB:1;             // 26
        UINT RDTSCP:1;              // 27
        UINT Reserved4:1;           // 28
        UINT LM:1;                  // 29
        UINT _3DNOWEXT:1;           // 30
        UINT _3DNOW:1;              // 31
    } Intel;  // high order

    struct { // low order
        UINT FPU:1;                 // 0
        UINT VME:1;                 // 1
        UINT DE:1;                  // 2
        UINT PSE:1;                 // 3
        UINT TSC:1;                 // 4
        UINT MSR:1;                 // 5
        UINT PAE:1;                 // 6
        UINT MCE:1;                 // 7
        UINT CX8:1;                 // 8
        UINT APIC:1;                // 9
        UINT Reserved1:1;           // 10
        UINT SYCALL:1;              // 11
        UINT MTRR:1;                // 12
        UINT PGE:1;                 // 13
        UINT MCA:1;                 // 14
        UINT CMOV:1;                // 15
        UINT PAT:1;                 // 17
        UINT PSE36:1;               // 17
        UINT Reserved2:1;           // 18
        UINT MP:1;                  // 19
        UINT NX:1;                  // 20
        UINT Reserved3:1;           // 21
        UINT MMXEXT:1;              // 22
        UINT MMX:1;                 // 23
        UINT FXSR:1;                // 24
        UINT FXSROPT:1;             // 25
        UINT PDPE1GB:1;             // 26
        UINT RDTSCP:1;              // 27
        UINT Reserved4:1;           // 28
        UINT LM:1;                  // 29
        UINT _3DNOWEXT:1;           // 30
        UINT _3DNOW:1;              // 31
    } AMD;  // high order
    UINT w;
} EDXF8;


typedef union {
    struct registers
    {
        DWORD EAX;
        DWORD EBX;
        DWORD ECX;
        DWORD EDX;
    } named;

    unsigned __int32 data[4];

    struct registersF1
    {
        EAXF1 eax;
        EBXF1 ebx;
        ECXF1 ecx;
        EDXF1 edx;
    } regF1;

    struct registersF7
    {
        EAXF1 eax;
        EBXF7 ebx;
        ECXF7 ecx;
        EDXF1 edx;
    } regF7;

    struct registersF8
    {
        EAXF1 eax;
        EBXF1 ebx;
        ECXF8 ecx;
        EDXF8 edx;
    } regF8;
} CPUID_Info;