//////////////////////////////////////////////////////////////////////////////
//   Author:       George Potts
//   Copyright:    PottsSoftware
//   Initial Date: 2015/08/13
// 
//   Description:  Implements many of the operations for 8-bit variables.
//
//   Notes:
//
//   Revision History:
//
//////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include "SSELibrary.h"
#include <immintrin.h>

#pragma region Abs_I8
SSELIBRARY_API void AVX2_Abs_I8(char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
	__m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
	{
		__m256i m_src = _mm256_load_si256(ptrSrc);
        __m256i result = _mm256_abs_epi8(m_src);
		_mm256_store_si256(ptrSrc, result);
		ptrSrc++;
	}
}

// SSSE3
SSELIBRARY_API void SSE_Abs_I8(char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrSrc = (__m128i*)src;

    for( ; loops > 0; loops--)
    {
        __m128i m_src = _mm_load_si128(ptrSrc);
        __m128i result = _mm_abs_epi8( m_src);
        _mm_store_si128(ptrSrc,result);
        ptrSrc++;
    }
}

SSELIBRARY_API void CPP_Abs_I8(char* src, size_t bytes)
{
	signed char * psrc = (signed char*)src;
    for (; bytes > 0; bytes--)
	{
        if (*psrc < 0) *psrc = 0 - *psrc;
        psrc++;
	}
}
#pragma endregion

#pragma region Sign_I8
SSELIBRARY_API void AVX2_Sign_I8(char * dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i * ptrDst = (__m256i*)dst;
    __m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
    {
        __m256i m_dst = _mm256_loadu_si256(ptrDst);
        __m256i m_src = _mm256_loadu_si256(ptrSrc);
        __m256i result = _mm256_sign_epi8(m_dst, m_src); // AVX2 intrinsic
        _mm256_store_si256(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

// SSSE3
// Negate packed 8-bit integers in dst when the corresponding signed 8-bit integer in src is negative, 
// and store the results in dst. Element in dst are zeroed out when the corresponding element in src is zero.
SSELIBRARY_API void SSSE3_Sign_I8(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;

    for (; loops > 0; loops--)
    {
        __m128i m_dst = _mm_load_si128(ptrDst);
        __m128i m_src = _mm_load_si128(ptrSrc);
        __m128i result = _mm_sign_epi8(m_dst, m_src);
        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

SSELIBRARY_API void CPP_Sign_I8(char* dst, char* src, size_t bytes)
{
	char * pdst = (char*)dst;
	char * psrc = (char*)src;
    for (; bytes > 0; bytes--)
	{
        if (*psrc < 0) *pdst = -*pdst;
        else if (*psrc == 0) *pdst = 0;
        psrc++;
        pdst++;
	}
}
#pragma endregion

#pragma region BlendV
SSELIBRARY_API void AVX2_BlendV_I8(char * dst, char* src, char* mask, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i * ptrDst = (__m256i*)dst;
    __m256i * ptrSrc = (__m256i*)src;
    __m256i * ptrMask = (__m256i*)mask;

    for (; loops > 0; loops--)
    {
        __m256i m_dst = _mm256_loadu_si256(ptrDst);
        __m256i m_src = _mm256_loadu_si256(ptrSrc);
        __m256i m_mask = _mm256_loadu_si256(ptrMask);
        __m256i result = _mm256_blendv_epi8(m_dst, m_src, m_mask); // AVX2 intrinsic
        _mm256_store_si256(ptrDst,result);
        ptrDst++;
        ptrSrc++;
        ptrMask++;
    }
}
// SSE 4.1
SSELIBRARY_API void SSE_BlendV_I8(char* dst, char* src, char * mask, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;
    __m128i * ptrMask = (__m128i*)mask;

    for( ; loops > 0; loops--)
    {
        __m128i m_dst = _mm_load_si128(ptrDst);
        __m128i m_src = _mm_load_si128(ptrSrc);
        __m128i m_mask = _mm_load_si128(ptrMask);
        __m128i result = _mm_blendv_epi8(m_dst, m_src, m_mask);
        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
        ptrMask++;
    }
}
SSELIBRARY_API void CPP_BlendV_I8(char* dst, char* src, char * mask, size_t bytes)
{
	char * pdst = (char*)dst;
	char * psrc = (char*)src;
	char * pmask = (char*)mask;
    for (; bytes > 0; bytes--)
	{
        if (*pmask & 0x80) *pdst = *psrc;
        psrc++;
        pdst++;
        pmask++;
	}
}
#pragma endregion

