//////////////////////////////////////////////////////////////////////////////
//   Author:       George Potts
//   Copyright:    PottsSoftware
//   Initial Date: 2015/08/14
// 
//   Description:  Implement the logical bitwise operations - AND, OR, XOR, ANDNOT
//
//   Notes:
//
//   Revision History:
//
//////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include "SSELibrary.h"
#include <immintrin.h>

#pragma region AND
// SSE2
SSELIBRARY_API void SSE_AND(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;

    for (; loops > 0; loops--)
    {
        __m128i m128_dst = _mm_load_si128(ptrDst);
        __m128i m128_src = _mm_load_si128(ptrSrc);
        __m128i result = _mm_and_si128(m128_dst, m128_src);
        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

// AVX1
SSELIBRARY_API void AVX_AND(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i * ptrDst = (__m256i*)dst;
    __m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
    {
        __m256i m_dst = _mm256_loadu_si256(ptrDst);
        __m256i m_src = _mm256_loadu_si256(ptrSrc);
        __m256d m_dst_d = _mm256_castsi256_pd(m_dst);   // generates no instructions
        __m256d m_src_d = _mm256_castsi256_pd(m_src);   // generates no instructions
        __m256d result = _mm256_and_pd(m_dst_d, m_src_d);
        __m256i result_i = _mm256_castpd_si256(result);   // generates no instructions
        _mm256_store_si256(ptrDst,result_i);
        ptrDst++;
        ptrSrc++;
    }
}

// AVX2
SSELIBRARY_API void AVX2_AND(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i * ptrDst = (__m256i*)dst;
    __m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
    {
        __m256i m_dst = _mm256_loadu_si256(ptrDst);
        __m256i m_src = _mm256_loadu_si256(ptrSrc);
        __m256i result = _mm256_and_si256(m_dst, m_src);
        _mm256_store_si256(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

SSELIBRARY_API void CPP_AND(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__int64) -1)/ sizeof(__int64);
    __int64* dPtr = (__int64*)dst;
    __int64* sPtr = (__int64*)src;

    for (; loops > 0; loops--)
    {
        *dPtr = *dPtr & *sPtr;
        dPtr++;
        sPtr++;
    }
}
#pragma endregion AND

#pragma region OR
// SSE2
SSELIBRARY_API void SSE_OR(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;

    for (; loops > 0; loops--)
    {
        __m128i m128_dst = _mm_load_si128(ptrDst);
        __m128i m128_src = _mm_load_si128(ptrSrc);
        __m128i result = _mm_or_si128(m128_dst, m128_src);
        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

// AVX1
SSELIBRARY_API void AVX_OR(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i * ptrDst = (__m256i*)dst;
    __m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
    {
        __m256i m_dst = _mm256_loadu_si256(ptrDst);
        __m256i m_src = _mm256_loadu_si256(ptrSrc);
        __m256d m_dst_d = _mm256_castsi256_pd(m_dst);   // generates no instructions
        __m256d m_src_d = _mm256_castsi256_pd(m_src);   // generates no instructions
        __m256d result_d = _mm256_or_pd(m_dst_d, m_src_d);
        __m256i result = _mm256_castpd_si256(result_d);   // generates no instructions

        _mm256_store_si256(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

// AVX2
SSELIBRARY_API void AVX2_OR(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i * ptrDst = (__m256i*)dst;
    __m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
    {
        __m256i m_dst = _mm256_loadu_si256(ptrDst);
        __m256i m_src = _mm256_loadu_si256(ptrSrc);
        __m256i result = _mm256_or_si256(m_dst, m_src); // AVX2 intrinsic
        _mm256_store_si256(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

SSELIBRARY_API void CPP_OR(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__int64) -1)/ sizeof(__int64);
    __int64* dPtr = (__int64*)dst;
    __int64* sPtr = (__int64*)src;

    for (; loops > 0; loops--)
    {
        *dPtr = *dPtr | *sPtr;
        dPtr++;
        sPtr++;
    }
}
#pragma endregion OR

#pragma region XOR
// SSE2
SSELIBRARY_API void SSE_XOR(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;

    for (; loops > 0; loops--)
    {
        __m128i m128_dst = _mm_load_si128(ptrDst);
        __m128i m128_src = _mm_load_si128(ptrSrc);
        __m128i result = _mm_xor_si128(m128_dst, m128_src);
        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

// AVX1
SSELIBRARY_API void AVX_XOR(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i * ptrDst = (__m256i*)dst;
    __m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
    {
        __m256i m_dst = _mm256_loadu_si256(ptrDst);
        __m256i m_src = _mm256_loadu_si256(ptrSrc);
        __m256d m_dst_d = _mm256_castsi256_pd(m_dst);   // generates no instructions
        __m256d m_src_d = _mm256_castsi256_pd(m_src);   // generates no instructions
        __m256d result = _mm256_xor_pd(m_dst_d, m_src_d);
        __m256i result_i = _mm256_castpd_si256(result);   // generates no instructions
        _mm256_store_si256(ptrDst,result_i);
        ptrDst++;
        ptrSrc++;
    }
}

// AVX2
SSELIBRARY_API void AVX2_XOR(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i * ptrDst = (__m256i*)dst;
    __m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
    {
        __m256i m_dst = _mm256_loadu_si256(ptrDst);
        __m256i m_src = _mm256_loadu_si256(ptrSrc);
        __m256i result = _mm256_xor_si256(m_dst, m_src); // AVX2 intrinsic
        _mm256_store_si256(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}


SSELIBRARY_API void CPP_XOR(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__int64) -1)/ sizeof(__int64);
    __int64* dPtr = (__int64*)dst;
    __int64* sPtr = (__int64*)src;

    for (; loops > 0; loops--)
    {
        *dPtr = *dPtr ^ *sPtr;
        dPtr++;
        sPtr++;
    }
}
#pragma endregion XOR

#pragma region ANDNOT
// Bitwise (NOT dst) AND (src)

// SSE2
SSELIBRARY_API void SSE_ANDNOT(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i * ptrSrc = (__m128i*)src;

    for (; loops > 0; loops--)
    {
        __m128i m128_dst = _mm_load_si128(ptrDst);
        __m128i m128_src = _mm_load_si128(ptrSrc);
        __m128i result = _mm_andnot_si128(m128_dst, m128_src);
        _mm_store_si128(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

// AVX1
SSELIBRARY_API void AVX_ANDNOT(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i * ptrDst = (__m256i*)dst;
    __m256i * ptrSrc = (__m256i*)src;

    __m256i ONESi = _mm256_set1_epi8(-1);
    __m256d ONES = _mm256_castsi256_pd(ONESi);

    for (; loops > 0; loops--)
    {
        __m256i m_dst = _mm256_loadu_si256(ptrDst);
        __m256i m_src = _mm256_loadu_si256(ptrSrc);
        __m256d m_dst_d = _mm256_castsi256_pd(m_dst);   // generates no instructions
        __m256d m_src_d = _mm256_castsi256_pd(m_src);   // generates no instructions

        __m256d res_i = _mm256_xor_pd(m_dst_d, ONES);
        res_i = _mm256_and_pd(res_i, m_src_d);
        __m256i result = _mm256_castpd_si256(res_i);   // generates no instructions
        _mm256_store_si256(ptrDst,result);

        ptrDst++;
        ptrSrc++;
    }
}

// AVX2
SSELIBRARY_API void AVX2_ANDNOT(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i * ptrDst = (__m256i*)dst;
    __m256i * ptrSrc = (__m256i*)src;

    for (; loops > 0; loops--)
    {
        __m256i m_dst = _mm256_loadu_si256(ptrDst);
        __m256i m_src = _mm256_loadu_si256(ptrSrc);
        __m256i result = _mm256_andnot_si256(m_dst, m_src); // AVX2 intrinsic
        _mm256_store_si256(ptrDst,result);
        ptrDst++;
        ptrSrc++;
    }
}

SSELIBRARY_API void CPP_ANDNOT(char* dst, char* src, size_t bytes)
{
    size_t loops = (bytes + sizeof(__int64) -1)/ sizeof(__int64);
    __int64* dPtr = (__int64*)dst;
    __int64* sPtr = (__int64*)src;

    for (; loops > 0; loops--)
    {
        *dPtr = ~(*dPtr) & *sPtr;
        dPtr++;
        sPtr++;
    }
}
#pragma endregion ANDNOT

#pragma region MEMSET8
// SSE2
SSELIBRARY_API void SSE_MEMSET8(char* dst, size_t bytes, int val)
{
    size_t loops = (bytes + sizeof(__m128i) - 1) / sizeof(__m128i);
    __m128i * ptrDst = (__m128i*)dst;
    __m128i mVal = _mm_set1_epi8((char)val);

    for (; loops > 0; loops--)
    {
        _mm_store_si128(ptrDst,mVal);
        ptrDst++;
    }
}

// AVX1
SSELIBRARY_API void AVX_MEMSET8(char* dst, size_t bytes, int val)
{
    size_t loops = (bytes + sizeof(__m256i) - 1) / sizeof(__m256i);
    __m256i * ptrDst = (__m256i*)dst;
    __m256i mVal = _mm256_set1_epi8((char)val);

    for (; loops > 0; loops--)
    {
        _mm256_store_si256(ptrDst,mVal);
        ptrDst++;
    }
}

SSELIBRARY_API void CPP_MEMSET8(char* dst, size_t bytes, int val)
{
    size_t loops = (bytes + sizeof(__int64) -1)/ sizeof(__int64);
    __int64 mVal;
    __int64* dPtr = (__int64*)dst;

    memset(&mVal,val,sizeof(mVal));
    for (; loops > 0; loops--)
    {
        *dPtr = mVal;
        dPtr++;
    }
}

#pragma endregion MEMSET8